#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System.Collections.Generic;
using System.Linq;
using System.Threading;

namespace HeuristicLab.Core.Networks {
  [Item("ProgrammableNode", "A programmable node of a network.")]
  [StorableClass]
  public class ProgrammableNode : ProgrammableNetworkItem, INode {
    protected override string CodeTemplate {
      get { return ReadCodeTemplate("HeuristicLab.Optimization.Networks.Core.Networks.ProgrammableNodeTemplate.cs"); }
    }

    new public INetwork Parent {
      get { return (INetwork)base.Parent; }
      set {
        // NOTE: never call setter directly as the Parent property is set by the network which contains the node
        base.Parent = value;
      }
    }
    public override IEnumerable<INetworkItem> Children {
      get { return base.Children.Concat(Ports.AsEnumerable<INetworkItem>()); }
    }

    [Storable]
    private PortCollection ports;
    protected PortCollection Ports {
      get { return ports; }
    }
    private ReadOnlyKeyedItemCollection<string, IPort> readOnlyPorts;
    IKeyedItemCollection<string, IPort> INode.Ports {
      get {
        if (readOnlyPorts == null) readOnlyPorts = ports.AsReadOnly();
        return readOnlyPorts;
      }
    }

    [StorableConstructor]
    private ProgrammableNode(bool deserializing) : base(deserializing) { }
    private ProgrammableNode(ProgrammableNode original, Cloner cloner)
      : base(original, cloner) {
      ports = cloner.Clone(original.ports);
      foreach (var p in Ports)
        p.Parent = this;
      readOnlyPorts = null;
      RegisterPortsEvents();
    }
    public ProgrammableNode()
      : base("ProgrammableNode") {
      ports = new PortCollection();
      readOnlyPorts = null;
      RegisterPortsEvents();
    }
    public ProgrammableNode(string name)
      : base(name) {
      ports = new PortCollection();
      readOnlyPorts = null;
      RegisterPortsEvents();
    }
    public ProgrammableNode(string name, string description)
      : base(name, description) {
      ports = new PortCollection();
      readOnlyPorts = null;
      RegisterPortsEvents();
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      foreach (var p in Ports)
        p.Parent = this;
      RegisterPortsEvents();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ProgrammableNode(this, cloner);
    }

    #region Ports Events
    protected virtual void RegisterPortsEvents() {
      ports.ItemsAdded += Ports_ItemsAdded;
      ports.ItemsRemoved += Ports_ItemsRemoved;
      ports.ItemsReplaced += Ports_ItemsReplaced;
      ports.CollectionReset += Ports_CollectionReset;
    }
    protected virtual void Ports_ItemsAdded(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      foreach (var p in e.Items)
        p.Parent = this;
    }
    protected virtual void Ports_ItemsRemoved(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      foreach (var p in e.Items)
        p.Parent = null;
    }
    protected virtual void Ports_ItemsReplaced(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      foreach (var p in e.OldItems)
        p.Parent = null;
      foreach (var p in e.Items)
        p.Parent = this;
    }
    protected virtual void Ports_CollectionReset(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      foreach (var p in e.OldItems)
        p.Parent = null;
      foreach (var p in e.Items)
        p.Parent = this;
    }
    #endregion

    #region CompiledProgrammableNode
    [Item("CompiledProgrammableNode", "Abstract base class for compiled programmable nodes of a network.")]
    public abstract class CompiledProgrammableNode : CompiledProgrammableNetworkItem, INode {
      protected new ProgrammableNode Context {
        get { return (ProgrammableNode)base.Context; }
      }

      protected CompiledProgrammableNode(CompiledProgrammableNode original, Cloner cloner)
        : base(original, cloner) {
      }
      protected CompiledProgrammableNode(ProgrammableNode context)
        : base(context) {
      }

      #region INode Members
      public new INetwork Parent {
        get { return Context.Parent; }
        set {
          // NOTE: never call setter directly as the Parent property is set by the network which contains the node
          Context.Parent = value;
        }
      }

      public IKeyedItemCollection<string, IPort> Ports {
        get { return Context.Ports; }
      }
      #endregion

      #region Context Events
      protected override void RegisterContextEvents() {
        base.RegisterContextEvents();
        Context.Ports.ItemsAdded += Ports_ItemsAdded;
        Context.Ports.ItemsRemoved += Ports_ItemsRemoved;
        Context.Ports.ItemsReplaced += Ports_ItemsReplaced;
        Context.Ports.CollectionReset += Ports_CollectionReset;
        foreach (var p in Context.Ports)
          RegisterPortEvents(p);
      }
      protected override void DeregisterContextEvents() {
        foreach (var p in Context.Ports)
          DeregisterPortEvents(p);
        Context.Ports.ItemsAdded -= Ports_ItemsAdded;
        Context.Ports.ItemsRemoved -= Ports_ItemsRemoved;
        Context.Ports.ItemsReplaced -= Ports_ItemsReplaced;
        Context.Ports.CollectionReset -= Ports_CollectionReset;
        base.DeregisterContextEvents();
      }
      private void Ports_ItemsAdded(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
        foreach (var p in e.Items)
          RegisterPortEvents(p);
      }
      private void Ports_ItemsRemoved(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
        foreach (var p in e.Items)
          DeregisterPortEvents(p);
      }
      private void Ports_ItemsReplaced(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
        foreach (var p in e.OldItems)
          DeregisterPortEvents(p);
        foreach (var p in e.Items)
          RegisterPortEvents(p);
      }
      private void Ports_CollectionReset(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
        foreach (var p in e.OldItems)
          DeregisterPortEvents(p);
        foreach (var p in e.Items)
          RegisterPortEvents(p);
      }
      protected virtual void RegisterPortEvents(IPort port) {
        var mp = port as IMessagePort;
        if (mp != null)
          mp.MessageReceived += MessagePort_MessageReceived;
      }
      protected virtual void DeregisterPortEvents(IPort port) {
        var mp = port as IMessagePort;
        if (mp != null)
          mp.MessageReceived -= MessagePort_MessageReceived;
      }
      protected virtual void MessagePort_MessageReceived(object sender, EventArgs<IMessage, CancellationToken> e) { }
      #endregion
    }
    #endregion
  }
}
